/*
    Tile Studio Scroll Demo Example

    This program demonstrates how Tile Studio can be used with
    DJGPP/Allegro.

    Mike Wiering (mike.wiering@cs.kun.nl)
*/

#include <stdlib.h>
#include <stdio.h>

#include "allegro.h"

#define DATA_FILE  "scroll.dat"

#define SCREEN_WIDTH   640
#define SCREEN_HEIGHT  480
#define BITS_PER_PIXEL  16

#include "gfx\ScrollGfx.c"  /* generated by Tile Studio */

int LayerCount = 0;

int XView = 0;
int YView = 0;
int XScrollSpeed = 0;
int YScrollSpeed = 0;

int iMaxXScrollSpeed = 4;
int iMaxYScrollSpeed = 4;
int iActualXView = 0;
int iActualYView = 0;

DATAFILE * datafile;
BITMAP *buffer;


typedef struct RECT
{
  int top, left, right, bottom;
} RECT;

#define MAX_BITMAP 32

typedef struct BITMAP_INFO
{
  BITMAP * BMP;
  int iBitmapWidth;
  int iBitmapHeight;
  int iBlockWidth;
  int iBlockHeight;
  int iBlockCountX;
  int iBlockCountY;
  int iSequences;
  int *ipSequenceLength;
  int *ipSequencePosition;
  int *ipSequenceCounter;
  int **ippSequenceData;
} BITMAP_INFO;

static BITMAP_INFO biBitmap[MAX_BITMAP];


static int InitBitmap (const struct GFX_BITMAP_DATA * gfxBitmapData)
{
  int w = gfxBitmapData->BitmapWidth;
  int h = gfxBitmapData->BitmapHeight;
  int index = gfxBitmapData->BitmapIdentifier;

  BITMAP_INFO *bip = &biBitmap[0];
  int id = 0;
  while (bip->BMP)
  {
    bip++;
    id++;
    if (id >= MAX_BITMAP)
      return -1;
  }

  bip->BMP = create_bitmap (w, h);
  
  blit (datafile[index].dat, bip->BMP, 0, 0, 0, 0, w, h);

  bip->iBitmapWidth = gfxBitmapData->BitmapWidth;
  bip->iBitmapHeight = gfxBitmapData->BitmapHeight;
  bip->iBlockWidth = gfxBitmapData->TileWidth;
  bip->iBlockHeight = gfxBitmapData->TileHeight;
  bip->iBlockCountX = gfxBitmapData->TileCountX;
  bip->iBlockCountY = gfxBitmapData->TileCountY;
  bip->iSequences = gfxBitmapData->Sequences;
  bip->ippSequenceData = gfxBitmapData->SequenceData;
  bip->ipSequenceLength = gfxBitmapData->SequenceLength;
  bip->ipSequencePosition = gfxBitmapData->SequencePosition;
  bip->ipSequenceCounter = gfxBitmapData->SequenceCounter;
  return id;
}


static int BitmapDone (int id)
{
  BITMAP_INFO *bip = &biBitmap[id];

  if (bip->BMP)
    destroy_bitmap (bip->BMP);
  bip->BMP = NULL;
  return 0;
}


int RunSequences ()
{
  int id;
  for (id = 0; id < MAX_BITMAP; id++)
  {
    BITMAP_INFO *bip = &biBitmap[id];
    if (bip->BMP)
    {
      int i;
      for (i = 0; i < bip->iSequences; i++)
        if (--bip->ipSequenceCounter[i] <= 0)
        {
          if (++bip->ipSequencePosition[i] >= bip->ipSequenceLength[i])
            bip->ipSequencePosition[i] = 0;
          bip->ipSequenceCounter[i] =
              bip->ippSequenceData[i][2 * bip->ipSequencePosition[i] + 1];
        }
    }
  }
  return 0;
}


static unsigned short *sBoundMap = NULL;
static int BoundMapWidth = 0;
static int BoundMapHeight = 0;
static int BoundBlockWidth = 0;
static int BoundBlockHeight = 0;
static int TotalWidth = 0;
static int TotalHeight = 0;


static int InitBoundMap (unsigned short *map, int w, int h, int bid)
{
  BITMAP_INFO *bip = &biBitmap[bid];

  sBoundMap = map;
  BoundMapWidth = w;
  BoundMapHeight = h;
  BoundBlockWidth = bip->iBlockWidth;
  BoundBlockHeight = bip->iBlockHeight;
  TotalWidth = bip->iBlockWidth * w;
  TotalHeight = bip->iBlockHeight * h;

  XView = 0;
  YView = 0;

  return 0;
}

// move XView and YView towards iActualXView and iActualYView
int CorrectView ()
{
  if (iActualXView > XView)
    XView = MIN (XView + iMaxXScrollSpeed, iActualXView);
  if (iActualYView > YView)
    YView = MIN (YView + iMaxYScrollSpeed, iActualYView);
  if (iActualXView < XView)
    XView = MAX (XView - iMaxXScrollSpeed, iActualXView);
  if (iActualYView < YView)
    YView = MAX (YView - iMaxYScrollSpeed, iActualYView);

  if (XView > TotalWidth - SCREEN_WIDTH)
  {
    XView = TotalWidth - SCREEN_WIDTH;
    iActualXView = XView;
    XScrollSpeed = 0;
  }
  if (YView > TotalHeight - SCREEN_HEIGHT)
  {
    YView = TotalHeight - SCREEN_HEIGHT;
    iActualYView = YView;
    YScrollSpeed = 0;
  }

  if (XView < 0)
  {
    XView = 0;
    iActualXView = XView;
    XScrollSpeed = 0;
  }
  if (YView < 0)
  {
    YView = 0;
    iActualYView = YView;
    YScrollSpeed = 0;
  }
  return 0;
}

static int MoveScreenTo (int x, int y)
{
  XView = x;
  YView = y;
  iActualXView = XView;
  iActualYView = YView;

  CorrectView ();
  return 0;
}

#define MAX_LAYER 8

typedef struct LAYERMAPINFO
{
  int iBitmapID;
  int iMapWidth;
  int iMapHeight;
  signed short *sMap;
  int iScrollFunction;
  int iHShift;
  int iVShift;
} LAYERMAPINFO;

static LAYERMAPINFO lmiLayerMapInfo[MAX_LAYER];


static int InitLayerMap (int bitmapid,
                         int mapwidth, int mapheight,
                         signed short *map,
                         int scrollfunction,
                         int hshift,
                         int vshift)
{
  if (LayerCount < MAX_LAYER)
  {
    LAYERMAPINFO *lmip = &lmiLayerMapInfo[LayerCount];
    lmip->iBitmapID = bitmapid;
    lmip->sMap = map;
    lmip->iMapWidth = mapwidth;
    lmip->iMapHeight = mapheight;
    lmip->iScrollFunction = scrollfunction;
    lmip->iHShift = hshift;
    lmip->iVShift = vshift;
    LayerCount++;
    return 0;
  }
  else
    return -1;
}


int CalcLayerPos (int scrollfunction, int *x, int *y)
{
  switch (scrollfunction)
  {
    case 0:
    {
      *x = XView;
      *y = YView;
    }
    break;
    default:
    {
      *x = XView / scrollfunction;
      *y = YView / scrollfunction;
    }
  }
  return 0;
}


int DrawLayer (int layer)
{
  RECT r, src, dst;
  int xstart, ystart;
  int x, y;
  int w, h;
  int mapx, mapy;
  int mapw, maph;
  int curw, curh;
  int mappos, mapblock;
  int bmpx, bmpy;
  LAYERMAPINFO *lmip = &lmiLayerMapInfo[layer];

  if ((layer < LayerCount) && (lmip))
  {
    BITMAP_INFO *bip = &biBitmap[lmip->iBitmapID];

    if (bip)
    {
      w = bip->iBlockWidth;
      h = bip->iBlockHeight;

      mapw = lmip->iMapWidth;
      maph = lmip->iMapHeight;

      // calculate layer position
      CalcLayerPos (lmip->iScrollFunction, &xstart, &ystart);
      xstart += lmip->iHShift;
      ystart += lmip->iVShift;
      while (xstart < 0)
        xstart += mapw * w;
      while (ystart < 0)
        ystart += maph * h;

      mapy = ystart / h;
      r.top = ystart % h;

      for (y = 0; y < SCREEN_HEIGHT; )
      {
        mapx = xstart / w;
        r.left = xstart % w;
        curh = MIN (SCREEN_HEIGHT - y, h - r.top);

        for (x = 0; x < SCREEN_WIDTH; )
        {
          curw = MIN (SCREEN_WIDTH - x, w - r.left);

          // calculate position in map, read block number
          mappos = ((mapx % mapw) + ((mapy % maph) * mapw));
          mapblock = *(signed short *) &lmip->sMap[mappos];

          if (mapblock < 0)  // sequence ID, draw current frame
          {
            int seq = -mapblock - 1;
            if (seq < bip->iSequences)
            {
              int pos = bip->ipSequencePosition[seq];
              int *seqdata = bip->ippSequenceData[seq];
              mapblock = seqdata[2 * pos];
            }
            else
            {
              seq -= bip->iSequences;
              mapblock = 0;
            }
          }
          mapblock--;

          dst.left = x;
          dst.top = y;
          dst.right = x + curw;
          dst.bottom = y + curh;

          if (mapblock >= 0)
          {
            bmpx = (mapblock % bip->iBlockCountX) * w;
            bmpy = (mapblock / bip->iBlockCountX) * h;

            src.left = r.left + bmpx;
            src.top = r.top + bmpy;
            src.right = src.left + curw;
            src.bottom = src.top + curh;

            masked_blit (bip->BMP, buffer, src.left, src.top,
                              dst.left, dst.top, curw, curh);
          }
          
          x += curw;
          r.left = 0;
          mapx++;
        }

        y += curh;
        r.top = 0;
        mapy++;
      }
    }
  }
  return 0;
}



int InitBitmaps ()
{
  int id = 0;
  memset (&biBitmap, 0, sizeof (biBitmap));
  LayerCount = 0;

  id = InitBitmap (&GfxBackGrBitmap);
  InitLayerMap (id, GfxBackGrMap1MapWidth, GfxBackGrMap1MapHeight,
                &BackGrMap1MapData[0][0], 3, 0, 0);

  id = InitBitmap (&GfxBlocksBitmap);
  InitLayerMap (id, GfxBlocksLevel1MapWidth, GfxBlocksLevel1MapHeight,
                &BlocksLevel1MapData[0][0], 1, 0, 0);
  InitBoundMap (&BlocksLevel1BoundMapData[0][0],
                GfxBlocksLevel1MapWidth, GfxBlocksLevel1MapHeight, id);
  return 0;
}


int DrawFrame ()
{
  int x = XView / BoundBlockWidth;
  int y = YView / BoundBlockHeight;
  int w = SCREEN_WIDTH / BoundBlockWidth;
  int h = SCREEN_HEIGHT / BoundBlockHeight;

  DrawLayer (0);
  DrawLayer (1);

  RunSequences ();

  iActualXView += XScrollSpeed;
  iActualYView += YScrollSpeed;
  CorrectView ();

  vsync ();
  blit (buffer, screen, 0, 0, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);

  return TRUE;
}


int main (int argc, char *argv[])
{
  char buf[256];

  int Alpha;
  PALETTE pal;

  allegro_init ();
  install_keyboard ();

 // set_color_conversion (COLORCONV_REDUCE_TRUE_TO_HI);
  set_color_depth (BITS_PER_PIXEL);
  if (set_gfx_mode (GFX_AUTODETECT, SCREEN_WIDTH, SCREEN_HEIGHT, 0, 0) != 0)
  {
    allegro_exit ();
    printf ("Error setting graphics mode\n%s\n\n", allegro_error);
    return 1;
  }

  buffer = create_bitmap (SCREEN_WIDTH, SCREEN_HEIGHT);
  clear (buffer);

  clear_to_color (screen, makecol (0, 0, 0));

  generate_332_palette (pal);
  set_palette (pal);

  /* load the datafile into memory */
  replace_filename (buf, argv[0], DATA_FILE, sizeof (buf));
  datafile = load_datafile (buf);
  if (!datafile)
  {
    allegro_exit ();
    printf ("Error loading %s!\n\n", buf);
    return 1;
  }

  InitBitmaps ();

  XScrollSpeed = 0;
  YScrollSpeed = 0;
  CorrectView ();

  do
  {
    DrawFrame ();
    
    if (key[KEY_RIGHT]) if (XScrollSpeed < iMaxXScrollSpeed)  XScrollSpeed++;
    if (key[KEY_LEFT])  if (XScrollSpeed > -iMaxXScrollSpeed) XScrollSpeed--;
    if (key[KEY_DOWN])  if (YScrollSpeed < iMaxYScrollSpeed)  YScrollSpeed++;
    if (key[KEY_UP])    if (YScrollSpeed > -iMaxYScrollSpeed) YScrollSpeed--;
  }
  while (!key[KEY_ESC]);
  
  allegro_exit ();
}

